/*\
//	BMP.c	( Windows 3.0 uncompressed format only )
//		for Hybrid coerce	4/29/1999	by Rick
//
//	based on
 *
 * ppmtobmp.c - Converts from a PPM file to a Microsoft Windows or OS/2
 * .BMP file.
 *
 * Copyright (C) 1992 by David W. Sanderson.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  This software is provided "as is"
 * without express or implied warranty.
 *
 * Revision 1.9  1992/11/24  19:39:33  dws
 *
//
*/

#include	"defs.h"
#include	"BMP.h"
#include	"pislice.h"

#define	MAXCOLORS 256


/* prototypes */
extern int	loadBMPfile (picsliceinfo*);
extern int	drawBMP (picsliceinfo*);
extern void	EndOfDraw (picsliceinfo*);

static int	littlelong (int a);
static short	littleshort (short a);
static int	BMPwritefileheader (FILE *fp, unsigned long bitcount,
									unsigned long x, unsigned long y);
static int	BMPwritergbtable (FILE *fp, int bpp, int colors,
				unsigned short *R, unsigned short *G, unsigned short *B);
static void	BMPwriterow (unsigned char *line, void *row,
				 unsigned long cx, unsigned short bpp);
static int	BMPwritebits (FILE *fp, unsigned long cx, unsigned long cy,
					unsigned short cBitCount, void *pixels);
static int	colorstobpp (int colors);
static void	BMPEncode (FILE *fp, int x, int y, void *pixels,
    int colors, unsigned short *R, unsigned short *G, unsigned short *B);


static int
littlelong (a)
int a;
{
#ifdef BIGENDIAN
	return ((a & 0xff) << 24)
		 + (((a >> 8) & 0xff) << 16)
		 + (((a >> 16) & 0xff) << 8)
		 +  ((a >> 24) & 0xff);
#else
	return a;
#endif
}

static short
littleshort (a)
short a;
{
#ifdef BIGENDIAN
	return ((a & 0xff) << 8)
			+ ((a >> 8) & 0xff);
#else
	return a;
#endif
}

/*
 * BMP writing
 */

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritefileheader (fp, bitcount, x, y)
FILE *fp;
unsigned long bitcount;
unsigned long x;
unsigned long y;
{
	BMPHEADER header;

	header.file.head1 = 'B';
	header.file.head2 = 'M';

	/* cbSize */
	header.file.cbSize = littlelong (BMPlenfile (bitcount, x, y));

	/* xHotSpot */
	header.file.xHotSpot = 0;

	/* yHotSpot */
	header.file.yHotSpot = 0;

	/* offBits */
	header.file.offBits = littlelong (BMPoffbits (bitcount));

	/* cbFix */
	header.info.cbFix = littlelong (40);

	/* cx */
	header.info.cx = littlelong (x);
	/* cy */
	header.info.cy = littlelong (y);
	/* cPlanes */
	header.info.cPlanes = littleshort (1);
	/* cBitCount */
	header.info.cBitCount = littleshort ((short)bitcount);

	/*
	 * We've written 16 bytes so far, need to write 24 more
	 * for the required total of 40.
	*/

	header.info.filler[0] = 0;
	header.info.filler[1] = 0;
	header.info.filler[2] = 0;
	header.info.filler[3] = 0;
	header.info.filler[4] = 0;
	header.info.filler[5] = 0;

	fwrite (&header, sizeof(BMPHEADER), 1, fp);

	return sizeof (BMPHEADER);
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritergbtable (fp, bpp, colors, R, G, B)
FILE *fp;
int bpp;
int colors;
unsigned short *R;
unsigned short *G;
unsigned short *B;
{
	int i;
	long ncolors;
	RGBQUAD rgbtable[MAXCOLORS];

	for (i = 0; i < colors; i++)
	{
		rgbtable[i].b = (unsigned char)(B[i]);
		rgbtable[i].g = (unsigned char)(G[i]);
		rgbtable[i].r = (unsigned char)(R[i]);
		rgbtable[i].filler = 0;
	}

	ncolors = (1 << bpp);

	for (; i < ncolors; i++)
	{
		rgbtable[i].b = 0;
		rgbtable[i].g = 0;
		rgbtable[i].r = 0;
		rgbtable[i].filler = 0;
	}

	fwrite (rgbtable, sizeof(RGBQUAD), ncolors, fp);

	return ncolors * 4;
}

static void
BMPwriterow (line, row, cx, bpp)
unsigned char *line;
void *row;
unsigned long cx;
unsigned short bpp;
{
	unsigned int nbyte = 0;
	unsigned int x;

	switch (bpp)
	{
	case 1:{
		unsigned short *w = (unsigned short*)line;
		unsigned short b;
		unsigned char *p = row;
		char nowbit = 16;

		for (x = 0; x < cx; x++)
		{
			if (--nowbit < 0)
			{
				*w++ = b;
				nowbit = 16 - 1;
			}
			b *= 2;
			b += *p++;
		}
		if (nowbit)	b= b << nowbit;
		*w++ = b;
		nbyte = (cx + 7) >> 3;
		break;
	}
	case 4:{
		unsigned short *w = (unsigned short*)line;
		unsigned short b;
		unsigned char *p = row;
		char nowbit = 4;

		for (x = 0; x < cx; x++)
		{
			if (--nowbit < 0)
			{
				*w++ = b;
				nowbit = 4 - 1;
			}
			b = b << 4;
			b += *p++;
		}
		if (nowbit)	b = b << (nowbit << 2);
		*w++ = b;
		nbyte = (cx + 1) >> 1;
		break;
	}
	case 8:{
		unsigned char *w= line;
		unsigned char *p= row;

		for (x = 0; x < cx; x++)	*w++ = *p++;
		nbyte = cx;
		break;
	}
	case 15:{
		unsigned char *w= line;
		unsigned short *p= row;

		for (x = 0; x < cx; x++, p++)
		{
			unsigned char b;

			b = *p & 0x1f;			/* R */
			*w++ = (b << 3) | (b >> 2);
			b = *p >> 10;			/* G */
			*w++ = (b << 3) | (b >> 2);
			b = (*p >> 5) & 0x1f;	/* B */
			*w++ = (b << 3) | (b >> 2);
		}
		nbyte= cx * 3;
		break;
	}
	default:{
		unsigned char *w = line;
		unsigned char *p = row;

		for (x = 0; x < cx; x++)
		{
			unsigned char r, g, b;

			r = *p++;
			g = *p++;
			b = *p++;
			*w++ = b;
			*w++ = g;
			*w++ = r;
		}
		nbyte = cx * 3;
		break;
	}
	}

	/*
	 * Make sure we write a multiple of 4 bytes.
	*/
	while (nbyte & 3)	line[nbyte++] = 0;
}

/*
 * returns the number of bytes written, or -1 on error.
 */
static int
BMPwritebits (fp, cx, cy, cBitCount, pixels)
FILE *fp;
unsigned long cx;
unsigned long cy;
unsigned short cBitCount;
void *pixels;
{
	long y;
	int cols = BMPlenline (cBitCount, cx);
	unsigned char *line = (unsigned char*)MALLOC (cols);

	if ((int)line < 0)	return -1;

	/*
	 * The picture is stored bottom line first, top line last
	 */

	for (y = cy - 1; y >= 0; y--)
	{
		unsigned char *row;

		row = pixels;
		if (InBpp == 2)
		{
			row += y * cx * 2;
		}
		else if (InBpp == 4)
		{
			row += y * cx * 3;
		}
		else
		{
			row += y * cx;
		}
		BMPwriterow (line, row, cx, (InBpp == 2)? 15:cBitCount);
		fwrite (line, sizeof(unsigned char), cols, fp);
	}

	MFREE (line);
	return cols * cy;
}

/*
 * Return the number of bits per pixel required to represent the
 * given number of colors.
 */

static int
colorstobpp (colors)
int colors;
{
	int bpp = 24;

	if (!colors)
	{
		error ("can't have less than one color");
	}
	else if (colors <= 2)	bpp = 1;
	else if (colors <= 16)	bpp = 4;
	else if (colors <= 256)	bpp = 8;
	else if ((colors >> 24) > 1)
	{
		error ("too many colors");
	}

	return bpp;
}

/*
 * Write a BMP file of the given class.
 *
 * Note that we must have 'colors' in order to know exactly how many
 * colors are in the R, G, B, arrays.  Entries beyond those in the
 * arrays are undefined.
 */
static void
BMPEncode (fp, x, y, pixels, colors, R, G, B)
FILE *fp;
int x;
int y;
void *pixels;
int colors;		/* number of valid entries in R,G,B */
unsigned short *R;
unsigned short *G;
unsigned short *B;
{
	int bpp;	/* bits per pixel */
	unsigned long nbyte = 0;

	bpp = colorstobpp (colors);

	/*
	 * I have found empirically at least one BMP-displaying program
	 * that can't deal with (for instance) using 3 bits per pixel.
	 * I have seen no programs that can deal with using 3 bits per
	 * pixel.  I have seen programs which can deal with 1, 4, and
	 * 8 bits per pixel.
	 *
	 * Based on this, I adjust actual the number of bits per pixel
	 * as follows.  If anyone knows better, PLEASE tell me!
	*/
	switch (bpp)
	{
/*	case 1:
		MaxColorGuaranteed = 2;
		break;/*
	case 2:
	case 3:
		bpp = 4;
	case 4:
/*		MaxColorGuaranteed = 16;*/
		break;
	case 5:
	case 6:
	case 7:
		bpp = 8;
/*		MaxColorGuaranteed = 65536;*/
		break;
	}

	nbyte += BMPwritefileheader (fp, bpp, x, y);
	if (bpp != 24)
	{
		nbyte += BMPwritergbtable (fp, bpp, colors, R, G, B);
	}

	if (nbyte != (BMPlenfileheader
				+ BMPleninfoheader
				+ BMPlenrgbtable (bpp)))
	{
		error ("abnormal data");
	}

	nbyte += BMPwritebits (fp, x, y, bpp, pixels);
		if (nbyte != BMPlenfile (bpp, x, y))
		{
			error ("BMPEncode failed");
		}
}


void
BMPSave (void)
{
	if (fp == NULL && (fp = fopen (File, "wb")) == NULL)
	{
		error ("Cannot create");
	}

	/* All set, let's do it. */
	BMPEncode (fp, Width, Height, Screen, MaxColorGuaranteed,
				Red, Green, Blue);

	if (fp != stdout)
	{
		fclose (fp);
	}
}

void
BMPLoad (void)
{
	int ret;
	picsliceinfo psi;

	if (fp)
	{
		psi.filehandle = 0;
	}
	else if ((psi.filehandle = OPEN (File, 0)) < 0)
	{
		error ("No such file");
	}
	if ((ret = loadBMPfile (&psi)) != PICSLICE_NORMAL)
	{
		if (ret == PICSLICE_NOMEMORY)
		{
			error ("Insufficient memory");
		}
		else if (ret == PICSLICE_EFORMAT)
		{
			error ("Is not bmp format");
		}
		else
		{
			error ("load error");
		}
	}
	Width = psi.width;
	Height = psi.height;
	MaxColorGuaranteed = psi.realwidth;

	/*	AllocateScreen()	*/
	Size = Width * Height * 2;
	if (Screen)	MFREE (Screen);
	if (InBpp == 4 && OutBpp != 2)
	{
		Screen = (int16*)MALLOC (Width * Height * 3);
	}
	else
	{
		Screen = (int16*)MALLOC (Size);
	}
	if ((int)Screen < 0)
	{
		EndOfDraw (&psi);
		error ("Insufficient memory");
	}

	(unsigned short*)psi.chainbuf = Screen;
	if ((ret = drawBMP (&psi)) != PICSLICE_TERMINATE)
	{
		error ("load error");
	}
}
